//-< DSYSTEM.H >-----------------------------------------------------*--------*
// D-Systemt                 Version 1.0         (c) 1999  GARRET    *     ?  *
// (Simple Distributed System)                                       *   /\|  *
//                                                                   *  /  \  *
//                          Created:     03-Apr-2013    K.A. Knizhnik  * / [] \ *
//                          Last update: 04-Apr-2013    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Distributed system classes
//-------------------------------------------------------------------*--------*

#ifndef __DSYSTEM_H__
#define __DSYSTEM_H__

#include <vector>
#include <exception>
#include "sockio.h"

typedef unsigned DSMsgCode;
typedef unsigned DSMsgSize;
typedef void const* DSMsgConstBody;
typedef void* DSMsgBody;
typedef unsigned DSNodeId;

struct DSMsgHeader 
{ 
    DSMsgCode code;
    DSMsgSize size;
};

/**
 * Exception throw in case of communcation errors
 */
class DSException : public std::exception
{
  public:
    int error;
    char const* message;

    virtual const char* what() const throw(); 
    DSException(char const* msg);    
};
    
/**
 * Class representing master side of distrubted system.
 * Master waits and accept connections from slave nodes (cents), send them equest messaqge, 
 * receive responses and perform syschronization.
 * Node can be a master and a slave at the same time.
 */
class DSMaster
{
  public:
    /**
     * Wait and establish connections with nClients
     * @param nClients number of expected client connections
     * @param port accept port (used both for local and global domain sockets)
     * @param mode bit mask of DSSocket::Mode bits
     * @param linger socket linger time (0: disable)
     */
    void accept(size_t nClients, int port, int mode = 0, int linger = 0);
    
    /**
     * Broadcaset message with specified code to all clients
     * @param code message code
     * @param body optional message body
     * @param size optional message body size
     */
    void broadcast(DSMsgCode code, DSMsgConstBody body = NULL, DSMsgSize size = 0);    
    
    
    /**
     * Broadcaset message with specified code to all clients
     * @param code message code
     * @param body optional message body
     * @param size optional message body size
     */
    template<class T>
    void broadcast(DSMsgCode code, std::vector<T> const& body) { 
        broadcast(code, &body[0], body.size()*sizeof(T));
    }
    
    /**
     * Broadcaset message with "request" code to all clients and wait "response"  message from all of them
     * @param request sent message code
     * @param response expected response message code
     */
    void barrier(DSMsgCode request, DSMsgCode response);

    /**
     * Send message to the particular client
     * @param node client node ID
     * @param code message code
     * @param body optional message body
     * @param size optional message body size
     */
    void send(DSNodeId node, DSMsgCode code, DSMsgConstBody data = NULL, DSMsgSize size = 0);

    /**
     * Send message to the particular client
     * @param node client node ID
     * @param code message code
     * @param body message body
     */
    template<class T>
    void send(DSNodeId node, DSMsgCode code, std::vector<T> const& body) { 
        send(node, code, &body[0], body.size()*sizeof(T));
    }

    /**
     * Receive response without body from the particular client
     * @param node client node ID
     * @param code expected response message code
     */
    void receive(DSNodeId node, DSMsgCode code);
    
    /**
     * Receive response with body from the particular client in the provided buffer.
     * Exception is thrown if the buffer is smaller than received message body
     * @param code expected response message code
     * @param buf buffer for the received message body
     * @param bufSize size of the buffer
     * return received message body size
     */
    DSMsgSize receive(DSNodeId node, DSMsgCode code, DSMsgBody buf, DSMsgSize bufSize);
    
    /**
     * Receive response with body from the particular client.
     * Exception is thrown if the buffer is smaller than received message body
     * @param code expected response message code
     * @param buf buffer to receive message body. If we be resized to actuall message size
     */
    template<class T>
    void receive(DSNodeId node, DSMsgCode code, std::vector<T>& buf) { 
        DSMsgHeader hdr;
        clients[node].read(&hdr, sizeof hdr);
        if (hdr.code != code) { 
            throw DSException("Unexpected response code");
        }
        buf.resize(hdr.size/sizeof(T));
        clients[node].read(&buf[0], hdr.size);
    }   
    
    /**
     * Wait for reponse from all clients
     * @param code expected response message code
     */
    void wait(DSMsgCode code);

    /**
     * Close connections with clients
     */
    void disconnect();

    size_t nNodes() const { 
        return clients.size();
    }

  private:
    std::vector<DSSocket> clients;
};

/**
 * Class represntin slave side of distributed system.
 * Slave establish connection with master, handle master requests and send responces to the master.
 * Node can be a master and a slave at the same time.
 */
class DSSlave
{
  public:
    /**
     * Message handlers passed in "process" method
     * @param code received message code
     * @param received message body (NULL if bodyy is empty)
     * @param received message body size
     * @param ctx user defined context
     * @return true if processing of messges should be continuted, false to terminate "process: method
     */
    typedef bool (*DSMsgHandler)(DSMsgCode code, DSMsgConstBody body, DSMsgSize size, void* ctx);

    /**
     * Establish connection with master
     * @param host name of master node, use "locahost" flor local domain socket conenction
     * @param port port listening by master
     * @param mode bit mask of DSSocket::Mode bits
     * @param nAttemtps number of attempts to establish connection
     * @param timeout interval in seconds between reconnect attempts
     * @param linger socket linger time (0: disable)
     */
    void connect(char const* host, int port, int mode = 0, int nAttempts = 10, int timeout = 1, int linger = 0);

    /**
     * Wait message from server and process them using provided handlers. Message code is used as 0-based  index
     * in the array of handlers. Processing is performed until false is returned by message handler
     * @param handlers array wiht message handlers
     * @param nHandlers number of message handlers
     * @param ctx user defined conext passed in invoked handler
     */
    void process(DSMsgHandler const* handlers, size_t nHandlers, void* ctx);

    /**
     * Wait message from server and process them using provided handlers. Message code is used as 0-based  index
     * in the array of handlers. Processing is performed until false is returned by message handler
     * @param handlers array wiht message handlers
     * @param nHandlers number of message handlers
     * @param ctx user defined conext passed in invoked handler
     */
    void process(std::vector<DSMsgHandler> const& handlers, void* ctx) { 
        return process(&handlers[0], handlers.size(), ctx);
    }

    /**
     * Send message to the master
     * @param code message code
     * @param body optional message body
     * @param size optional message body size
     */
    void send(DSMsgCode code, DSMsgConstBody data = NULL, DSMsgSize size = 0);

    /**
     * Send message to the master
     * @param code message code
     * @param body message body
     */
    template<class T>
    void send(DSMsgCode code, std::vector<T> const& body) { 
        send(code, &body[0], body.size()*sizeof(T));
    }

    /**
     * Receive message from master in the provided buffer.
     * Exception is thrown if the buffer is smaller than received message body
     * @param code OUT: received message code
     * @param buf buffer for the received message body
     * @param bufSize size of the buffer
     * return received message body size
     */
    DSMsgSize receive(DSMsgCode& code, DSMsgBody buf, DSMsgSize bufSize);
        
    /**
     * Receive message from master
     * @param code OUT: received message code
     * @param buf buffer for the received message body. It will be resized to the actual body size.
     */
    template<class T>
    void receive(DSMsgCode& code, std::vector<T>& buf)
    {
        DSMsgHeader hdr;
        master.read(&hdr, sizeof hdr);
        code = hdr.code;
        buf.resize(hdr.size/sizeof(T));
        master.read(&buf[0], hdr.size);
    }

    /**
     * Close connections with master
     */
    void disconnect();

  private:
    DSSocket master;
};

#endif
