//-< DSYSTEM.CPP >---------------------------------------------------*--------*
// D-Systemt                 Version 1.0         (c) 1999  GARRET    *     ?  *
// (Simple Distributed System)                                       *   /\|  *
//                                                                   *  /  \  *
//                          Created:     03-Apr-2013    K.A. Knizhnik  * / [] \ *
//                          Last update: 04-Apr-2013    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Distributed system classes
//-------------------------------------------------------------------*--------*


#include <stdio.h>
#include <errno.h>
#include "dsystem.h"

const size_t maxMessageLength = 256;

DSException::DSException(char const* msg)
: error(errno), message(msg) {}


const char* DSException::what() const throw() 
{ 
    static char buf[maxMessageLength];
    sprintf(buf, "%s: %d", message, error);
    return buf;
}


void DSMaster::accept(size_t nClients, int port, int mode, int linger)
{
    disconnect();
    clients.resize(nClients);
    DSSocket::accept(clients, port, mode, linger);
}

void DSMaster::broadcast(DSMsgCode code, DSMsgConstBody body, DSMsgSize size)
{
    DSMsgHeader hdr;
    hdr.code = code;
    hdr.size = size;
        
    for (size_t node = 0, nClients = clients.size(); node < nClients; node++) { 
        clients[node].write(&hdr, sizeof hdr);
        if (size > 0) { 
            clients[node].write(body, size);
        }
    }
}

void DSMaster::barrier(DSMsgCode request, DSMsgCode response)
{
    broadcast (request);
    wait(response);
}

void DSMaster::send(DSNodeId node, DSMsgCode code, DSMsgConstBody data, DSMsgSize size)
{
    DSMsgHeader hdr;
    hdr.code = code;
    hdr.size = size;
    clients[node].write(&hdr, sizeof hdr);
    if (size > 0)  {  
        clients[node].write(data, size);
    }
}

void DSMaster::receive(DSNodeId node, DSMsgCode code)
{
    DSMsgHeader hdr;
    clients[node].read(&hdr, sizeof hdr);
    if (hdr.code != code || hdr.size != 0) { 
        throw DSException("Unecpected response code");
    }
}

DSMsgSize DSMaster::receive(DSNodeId node, DSMsgCode code, DSMsgBody buf, DSMsgSize bufSize)
{
    DSMsgHeader hdr;
    clients[node].read(&hdr, sizeof hdr);
    if (hdr.code != code) { 
        throw DSException("Unexpected response code");
    }
    if (hdr.size > 0) {
        if (hdr.size > bufSize) {             
            throw DSException("Response too large");
        }
        clients[node].read(buf, hdr.size);
        return hdr.size;
    }
    return 0;
}
   
void DSMaster::wait(DSMsgCode code)
{
    DSMsgHeader hdr;
    for (size_t node = 0, nClients = clients.size(); node < nClients; node++) { 
        clients[node].read(&hdr, sizeof hdr);
        if (hdr.code != code || hdr.size != 0) { 
            throw DSException("Unecpected response code");
        }
    }
}

void DSMaster::disconnect()
{
    for (size_t node = 0, nClients = clients.size(); node < nClients; node++) { 
        clients[node].disconnect();
    }
    clients.clear();
}

void DSSlave::connect(char const* host, int port, int mode, int nAttempts, int timeout, int linger)
{
    master.connect(host, port, mode, nAttempts, timeout, linger);
}

void DSSlave::process(DSMsgHandler const* handlers, size_t nHandlers, void* ctx)
{
    DSMsgHeader hdr;
    std::vector<char> body;
    do { 
        master.read(&hdr, sizeof hdr);
        if (hdr.code > nHandlers) { 
            throw DSException("Too large message code");
        }
        if (hdr.size > 0) { 
            body.resize(hdr.size);
            master.read(&body[0], hdr.size);
        }
    } while (handlers[hdr.code](hdr.code, &body[0], hdr.size, ctx));
}

void DSSlave::send(DSMsgCode code, DSMsgConstBody body, DSMsgSize size)
{
    DSMsgHeader hdr;
    hdr.code = code;
    hdr.size = size;
    master.write(&hdr, sizeof hdr);
    if (size > 0)  {  
        master.write(body, size);
    }
}
    
DSMsgSize DSSlave::receive(DSMsgCode& code, DSMsgBody buf, DSMsgSize bufSize)
{
    DSMsgHeader hdr;
    master.read(&hdr, sizeof hdr);
    code = hdr.code;
    if (hdr.size > 0) { 
        if (hdr.size > bufSize) { 
            throw DSException("Response too large");
        }
        master.read(buf, hdr.size);
    }
    return hdr.size;
}

void DSSlave::disconnect()
{
    master.disconnect();
}
