//-< SOCKIO.CPP >----------------------------------------------------*--------*
// D-Systemt                 Version 1.0         (c) 1999  GARRET    *     ?  *
// (Simple Distributed System)                                       *   /\|  *
//                                                                   *  /  \  *
//                          Created:     03-Apr-2013    K.A. Knizhnik  * / [] \ *
//                          Last update: 04-Apr-2013    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Socket interface
//-------------------------------------------------------------------*--------*

#ifndef _WIN32
#ifndef HPUX11
#include <sys/select.h>
#endif
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#if defined(_AIX)
#include <strings.h>
#endif  /* _AIX */
#include <stddef.h>
#include <errno.h>

extern "C" {
#include <netdb.h>
}

#else // _WIN32

#ifdef USE_MFC
#include <afx.h>
#include <winsock2.h>
#else
#include <windows.h>
#endif
#include <errno.h>

#define close(x) closesocket(x)
#define sleep(x) Sleep((x)*1000)

#endif

#include "dsystem.h"


char const* DSSocket::UnixSocketDir = "/tmp/";

void DSSocket::accept(std::vector<DSSocket>& clients, int port, int mode, int linger)
{
    DSSocket localGate;
    DSSocket globalGate;
    int rc;
    fd_set events;
    size_t nClients = clients.size();

    // Create local socket
#ifndef _WIN32
    if (!(mode & GlobalDomainOnly)) { 
        sockaddr sock;
        sock.sa_family = AF_UNIX;    
        int len = offsetof(sockaddr,sa_data) + sprintf(sock.sa_data, "%sp%u", UnixSocketDir, port);
        
        unlink(sock.sa_data); // remove file if existed
        
        localGate.sd = socket(AF_UNIX, SOCK_STREAM, 0);
        if (localGate.sd < 0) { 
            throw DSException("Failed to create local socket");
        }    
        if (bind(localGate.sd, &sock, len) < 0) {
            throw DSException("Failed to bind local socket");
        }    
        if (listen(localGate.sd, nClients) < 0) {
            throw DSException("Failed to listen local socket");
        }    
    }
#endif
    if (!(mode & LocalDomainOnly)) { 
        // create global socket
        sockaddr_in sock_inet;
        sock_inet.sin_family = AF_INET;
        sock_inet.sin_addr.s_addr = htonl(INADDR_ANY);
        sock_inet.sin_port = htons(port);
        
        globalGate.sd = socket(AF_INET, SOCK_STREAM, 0);
        if (globalGate.sd < 0) { 
            throw DSException("Failed to create global socket");
        }    
        if (!(mode & DoNotReuseAddress)) { 
            int on = 1;
            if (setsockopt(globalGate.sd, SOL_SOCKET, SO_REUSEADDR, (char*)&on, sizeof on) < 0) { 
                throw DSException("Failed to set global socket option");
            }    
        }    
        if (bind(globalGate.sd, (sockaddr*)&sock_inet, sizeof(sock_inet)) < 0) {
            throw DSException("Failed to bind global socket");
        }    
        if (listen(globalGate.sd, nClients) < 0) {
            throw DSException("Failed to listen global socket");
        }    
    }
    FD_ZERO(&events);
    for (size_t i = 0; i < nClients; ) { 
        int maxFD = 0;
        if (localGate.sd >= 0) { 
            FD_SET(localGate.sd, &events);
            maxFD = localGate.sd;
        }
        if (globalGate.sd >= 0) { 
            FD_SET(globalGate.sd, &events);
            if (globalGate.sd > maxFD) { 
                maxFD = globalGate.sd;
            }
        }
        while ((rc = select(maxFD+1, &events, NULL, NULL, NULL)) < 0 && errno == EINTR);
        if (rc < 0) { 
            throw DSException("Failed to select gate");
        }    
        if (localGate.sd >= 0 && FD_ISSET(localGate.sd, &events)) { 
            while((clients[i].sd = ::accept(localGate.sd, NULL, NULL )) < 0 && errno == EINTR);
            if (clients[i].sd < 0) { 
                throw DSException("Failed to accept local connection");
            }    
            if (linger != 0) { 
                struct linger l;
                l.l_onoff = 1;
                l.l_linger = linger;
                if (setsockopt(clients[i].sd, SOL_SOCKET, SO_LINGER, (char*)&l, sizeof l) != 0) {
                    throw DSException("Failed to set linger option");
                }
            }   
            i += 1;
        }
        if (globalGate.sd >= 0 && FD_ISSET(globalGate.sd, &events)) { 
            while((clients[i].sd = ::accept(globalGate.sd, NULL, NULL )) < 0 && errno == EINTR);
            if (clients[i].sd < 0) { 
                throw DSException("Failed to accept remote connection");
            }    
            if (mode & NoDelay) { 
                int enabled = 1;
                if (setsockopt(clients[i].sd, IPPROTO_TCP, TCP_NODELAY, (char*)&enabled, sizeof enabled) != 0) {
                    throw DSException("Failed to set TCP_NODELAY option");
                }
            }
            if (linger != 0) { 
                struct linger l;
                l.l_onoff = 1;
                l.l_linger = linger;
                if (setsockopt(clients[i].sd, SOL_SOCKET, SO_LINGER, (char*)&l, sizeof l) != 0) {
                    throw DSException("Failed to set linger option");
                }
            }   
            i += 1;
        }
    }
}

void DSSocket::connect(char const* host, int port, int mode, int nAttempts, int timeout, int linger)
{
    int rc;
    while (true) { 
#ifndef _WIN32
        if ((!(mode & GlobalDomainOnly) && strcmp(host, "localhost") == 0) || (mode & LocalDomainOnly)) { 
            // connect UNIX socket
            sockaddr sock;
            sock.sa_family = AF_UNIX;
            int len = offsetof(sockaddr,sa_data) +
                sprintf(sock.sa_data, "%sp%u", UnixSocketDir, port);
            if ((sd = socket(sock.sa_family, SOCK_STREAM, 0)) < 0) { 
                throw DSException("Failed to create socket");
                break;
            }
            do { 
                rc = ::connect(sd, &sock, len);
            } while (rc < 0 && errno == EINTR);
            
            if (rc < 0) { 
                close(sd);
                sd = -1;
                if (errno == ENOENT || errno == ECONNREFUSED) {
                    if (--nAttempts >= 0) {
                        sleep(timeout);
                        continue;
                    }
                } 
                throw DSException("Failed to connect local socket");
            } else { 
                if (linger != 0) { 
                    struct linger l;
                    l.l_onoff = 1;
                    l.l_linger = linger;
                    if (setsockopt(sd, SOL_SOCKET, SO_LINGER, (char*)&l, sizeof l) != 0) {
                        close(sd);
                        sd = -1;
                        throw DSException("Failed to set linger option");
                    }
                }   
                return;
            }
        } 
        else 
#endif
        if (!(mode & LocalDomainOnly)) 
        { 
            sockaddr_in sock_inet;
            struct hostent* hp;
            if ((hp = gethostbyname(host)) == NULL || hp->h_addrtype != AF_INET) { 
                throw DSException("Failed to locate host");
            }
            sock_inet.sin_family = AF_INET;  
            sock_inet.sin_port = htons(port);
            
            for (int i = 0; hp->h_addr_list[i] != NULL; i++) { 
                memcpy(&sock_inet.sin_addr, hp->h_addr_list[i], sizeof sock_inet.sin_addr);
                if ((sd = ::socket(sock_inet.sin_family, SOCK_STREAM, 0)) < 0) { 
                    throw DSException("Failed to create socket");
                }
                do { 
                    rc = ::connect(sd, (sockaddr*)&sock_inet, sizeof(sock_inet));
                } while (rc < 0 && errno == EINTR);
            
                if (rc >= 0) { 
                    if (mode & NoDelay) { 
                        int enabled = 1;
                        if (setsockopt(sd, IPPROTO_TCP, TCP_NODELAY, (char*)&enabled, sizeof enabled) != 0) {
                            ::close(sd);
                            sd = -1;
                            throw DSException("Failed to set TCP_NODELAY option");
                        }
                    }
                    if (linger != 0) { 
                        struct linger l;
                        l.l_onoff = 1;
                        l.l_linger = linger;
                        if (setsockopt(sd, SOL_SOCKET, SO_LINGER, (char*)&l, sizeof l) != 0) {
                            ::close(sd);
                            sd = -1;
                            throw DSException("Failed to set linger option");
                        }
                    }   
                    return;
                }
                close(sd);
                sd = -1;
            }
            if (errno == ENOENT || errno == ECONNREFUSED) {
                if (--nAttempts >= 0) {
                    sleep(timeout);
                    continue;
                }
            } 
            throw DSException("Failed to connect local socket");
        }
    }  
}

void DSSocket::read(void* buf, size_t size)
{
    size_t offs = 0;
    int rc;
    while (offs < size) { 
        while ((rc = ::send(sd, (char*)buf + offs, size - offs, 0)) < 0 && errno == EINTR);
        if (rc <= 0) { 
            throw DSException("Failed to receive data");
        }
        offs += rc;
    }
}

void DSSocket::write(const void* buf, size_t size)
{
    size_t offs = 0;
    int rc;
    while (offs < size) { 
        while ((rc = ::recv(sd, (char*)buf + offs, size - offs, 0)) < 0 && errno == EINTR);
        if (rc <= 0) { 
            throw DSException("Failed to send data");
        }
        offs += rc;
    }
}

void DSSocket::disconnect()
{
    if (sd >= 0) { 
        if (::close(sd) < 0) { 
            throw DSException("Failed to close socket");
        }
        sd = -1;
    }
}

DSSocket::~DSSocket()
{
    disconnect();
}
